import fs from 'node:fs';
import { createRequire } from 'node:module';
import { parseVueRequest as parseVueRequest$1 } from '@vitejs/plugin-vue';

function resolveCompiler(root) {
  const compiler = tryResolveCompiler(root) || tryResolveCompiler();
  if (!compiler) {
    throw new Error(
      `Failed to resolve vue/compiler-sfc.
@vitejs/plugin-vue requires vue (>=3.2.25) to be present in the dependency tree.`
    );
  }
  return compiler;
}
function tryResolveCompiler(root) {
  const vueMeta = tryRequire("vue/package.json", root);
  if (vueMeta && vueMeta.version.split(".")[0] >= 3) {
    return tryRequire("vue/compiler-sfc", root);
  }
}
const _require = createRequire(import.meta.url);
function tryRequire(id, from) {
  try {
    return from ? _require(_require.resolve(id, { paths: [from] })) : _require(id);
  } catch {
  }
}

function pascalCase(str) {
  return str.replace(/(?:\b|_)[a-z]/g, (c) => c.toUpperCase()).replace(/[\W_]+/g, "");
}
function parseVueRequest(id) {
  return parseVueRequest$1(id);
}

function createCache(config) {
  const descriptorCache = /* @__PURE__ */ new Map();
  const compiler = resolveCompiler(config.root);
  return {
    getDescriptor(filename) {
      if (!descriptorCache.has(filename)) {
        const { descriptor, errors } = compiler.parse(
          fs.readFileSync(filename, "utf8").toString(),
          {
            filename,
            sourceMap: config.command === "build" ? !!config.build.sourcemap : true,
            sourceRoot: config.root
          }
        );
        if (errors.length > 0) {
          throw errors[0];
        }
        descriptorCache.set(filename, descriptor);
      }
      return descriptorCache.get(filename);
    },
    updateFileCache(filename, code) {
      const { descriptor, errors } = compiler.parse(code, { filename });
      if (errors.length > 0) {
        throw errors[0];
      }
      descriptorCache.set(filename, descriptor);
    },
    getNestedComponents(filename) {
      return this.getDescriptor(filename).customBlocks.filter(
        (block) => block.type === "component" && typeof block.attrs.name === "string"
      ).map((block) => pascalCase(block.attrs.name));
    },
    hasFile(filename) {
      return descriptorCache.has(filename);
    }
  };
}

function genComponentBlockCode(virtualFilename, component) {
  return `import ${component} from '${virtualFilename}/${component}.vue';
export default function(Comp) {
  if (!Comp.components) {
    Comp.components = {};
  }
  Comp.components[${JSON.stringify(component)}] = ${component};
}`;
}
function genExportsCode(virtualFilename, components, mainCode) {
  const codes = [mainCode, "\n"];
  for (const component of components) {
    codes.push(
      "export { default as ",
      component,
      " } from '",
      virtualFilename,
      "/",
      component,
      ".vue';\n"
    );
  }
  return codes.join("");
}

function vueNestedSFC() {
  const prefix = "virtual:vue-nested-sfc";
  let config;
  let cache;
  return {
    name: "vue-nested-sfc",
    configResolved(resolvedConfig) {
      config = resolvedConfig;
    },
    buildStart() {
      cache = createCache(config);
    },
    resolveId(id) {
      if (id.startsWith(prefix)) {
        return id;
      }
    },
    load(id) {
      if (!id.startsWith(prefix)) {
        return;
      }
      const [, filename, component] = id.slice(prefix.length).match(/^(.*)\/([^/]+)\.vue$/) || [];
      const descriptor = cache.getDescriptor(filename);
      const componentBlock = descriptor.customBlocks.find(
        (block) => block.type === "component" && typeof block.attrs.name === "string" && pascalCase(block.attrs.name) === component
      );
      if (!componentBlock) {
        return "";
      }
      return { code: componentBlock.content, map: componentBlock.map };
    },
    transform(code, id) {
      const request = parseVueRequest(id);
      if (request.filename.startsWith(prefix)) {
        return;
      }
      if (!request.query.vue && request.filename.endsWith(".vue")) {
        return genExportsCode(
          prefix + request.filename,
          cache.getNestedComponents(id),
          code
        );
      } else if (request.query.type === "component" && request.query.name) {
        return genComponentBlockCode(
          prefix + request.filename,
          pascalCase(request.query.name)
        );
      }
    },
    async handleHotUpdate({ modules, read, file, server }) {
      if (!cache.hasFile(file)) {
        return modules;
      }
      const affectedModules = new Set(
        modules.filter((m) => !/type=component/.test(m.url))
      );
      const prevDescriptor = cache.getDescriptor(file);
      cache.updateFileCache(file, await read());
      const nextDescriptor = cache.getDescriptor(file);
      if (prevDescriptor.customBlocks.length !== nextDescriptor.customBlocks.length) {
        const mainModule = server.moduleGraph.getModuleById(file);
        if (mainModule) {
          affectedModules.add(mainModule);
        }
      }
      for (const block of prevDescriptor.customBlocks) {
        if (block.type !== "component") {
          continue;
        }
        if (typeof block.attrs.name !== "string") {
          continue;
        }
        const name = pascalCase(block.attrs.name);
        const nextBlock = nextDescriptor.customBlocks.find(
          (nextBlock2) => nextBlock2.type === "component" && typeof nextBlock2.attrs.name === "string" && pascalCase(nextBlock2.attrs.name) === name
        );
        if (!nextBlock || nextBlock.attrs.name !== block.attrs.name) {
          const mainModule = server.moduleGraph.getModuleById(file);
          if (mainModule) {
            affectedModules.add(mainModule);
          }
        }
        if (!nextBlock || block.content === nextBlock.content) {
          continue;
        }
        const componentModule = server.moduleGraph.getModuleById(
          `${prefix}${file}/${name}.vue`
        );
        if (!componentModule) {
          continue;
        }
        affectedModules.add(componentModule);
        const blockModule = [...componentModule.importers].find(
          (m) => m.url.includes("type=component") && m.url.includes(`name=${nextBlock.attrs.name}`)
        );
        if (!blockModule) {
          continue;
        }
        affectedModules.add(blockModule);
      }
      return [...affectedModules];
    }
  };
}

export { vueNestedSFC as default };
